"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.branchIntersection = exports.flattenBranch = exports.isTransformationBranch = exports.compareBranches = exports.isBranchComparison = void 0;
const intersection_js_1 = require("../parse/ast/intersection.js");
const domains_js_1 = require("../utils/domains.js");
const errors_js_1 = require("../utils/errors.js");
const compose_js_1 = require("./compose.js");
const rules_js_1 = require("./rules/rules.js");
const isBranchComparison = (comparison) => comparison?.lBranches !== undefined;
exports.isBranchComparison = isBranchComparison;
const compareBranches = (lConditions, rConditions, state) => {
    const result = {
        lBranches: lConditions,
        rBranches: rConditions,
        lExtendsR: [],
        rExtendsL: [],
        equalities: [],
        distinctIntersections: []
    };
    const pairs = rConditions.map((condition) => ({
        condition,
        distinct: []
    }));
    lConditions.forEach((l, lIndex) => {
        let lImpliesR = false;
        const distinct = pairs.map((rPairs, rIndex) => {
            if (lImpliesR || !rPairs.distinct) {
                return null;
            }
            const r = rPairs.condition;
            const subresult = (0, exports.branchIntersection)(l, r, state);
            if ((0, compose_js_1.isDisjoint)(subresult)) {
                // doesn't tell us about any redundancies or add a distinct pair
                return null;
            }
            else if (subresult === l) {
                result.lExtendsR.push(lIndex);
                // If l is a subtype of the current r branch, intersections
                // with the remaining branches of r won't lead to distinct
                // branches, so we set a flag indicating we can skip them.
                lImpliesR = true;
                return null;
            }
            else if (subresult === r) {
                result.rExtendsL.push(rIndex);
                // If r is a subtype of the current l branch, it is removed
                // from pairsByR because future intersections won't lead to
                // distinct branches.
                rPairs.distinct = null;
                return null;
            }
            else if ((0, compose_js_1.isEquality)(subresult)) {
                // Combination of l and r subtype cases.
                result.equalities.push([lIndex, rIndex]);
                lImpliesR = true;
                rPairs.distinct = null;
                return null;
            }
            else if ((0, domains_js_1.hasDomain)(subresult, "object")) {
                // Neither branch is a subtype of the other, return
                // the result of the intersection as a candidate
                // branch for the final union
                return subresult;
            }
            return (0, errors_js_1.throwInternalError)(`Unexpected predicate intersection result of type '${(0, domains_js_1.domainOf)(subresult)}'`);
        });
        if (!lImpliesR) {
            for (let i = 0; i < pairs.length; i++) {
                if (distinct[i]) {
                    pairs[i].distinct?.push(distinct[i]);
                }
            }
        }
    });
    result.distinctIntersections = pairs.flatMap((pairs) => pairs.distinct ?? []);
    return result;
};
exports.compareBranches = compareBranches;
const isTransformationBranch = (branch) => "rules" in branch;
exports.isTransformationBranch = isTransformationBranch;
const flattenBranch = (branch, ctx) => {
    if ((0, exports.isTransformationBranch)(branch)) {
        const result = (0, rules_js_1.flattenRules)(branch.rules, ctx);
        if (branch.morph) {
            if (typeof branch.morph === "function") {
                result.push(["morph", branch.morph]);
            }
            else {
                for (const morph of branch.morph) {
                    result.push(["morph", morph]);
                }
            }
        }
        return result;
    }
    return (0, rules_js_1.flattenRules)(branch, ctx);
};
exports.flattenBranch = flattenBranch;
const rulesOf = (branch) => branch.rules ?? branch;
const branchIntersection = (l, r, state) => {
    const lRules = rulesOf(l);
    const rRules = rulesOf(r);
    const rulesResult = (0, rules_js_1.rulesIntersection)(lRules, rRules, state);
    if ("morph" in l) {
        if ("morph" in r) {
            if (l.morph === r.morph) {
                return (0, compose_js_1.isEquality)(rulesResult) || (0, compose_js_1.isDisjoint)(rulesResult)
                    ? rulesResult
                    : {
                        rules: rulesResult,
                        morph: l.morph
                    };
            }
            return state.lastOperator === "&"
                ? (0, errors_js_1.throwParseError)((0, intersection_js_1.writeImplicitNeverMessage)(state.path, "Intersection", "of morphs"))
                : {};
        }
        return (0, compose_js_1.isDisjoint)(rulesResult)
            ? rulesResult
            : {
                rules: (0, compose_js_1.isEquality)(rulesResult) ? l.rules : rulesResult,
                morph: l.morph
            };
    }
    if ("morph" in r) {
        return (0, compose_js_1.isDisjoint)(rulesResult)
            ? rulesResult
            : {
                rules: (0, compose_js_1.isEquality)(rulesResult) ? r.rules : rulesResult,
                morph: r.morph
            };
    }
    return rulesResult;
};
exports.branchIntersection = branchIntersection;
