"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.serializeCase = exports.serializeDefinitionIfAllowed = exports.flattenBranches = void 0;
const union_js_1 = require("../parse/ast/union.js");
const domains_js_1 = require("../utils/domains.js");
const errors_js_1 = require("../utils/errors.js");
const generics_js_1 = require("../utils/generics.js");
const objectKinds_js_1 = require("../utils/objectKinds.js");
const paths_js_1 = require("../utils/paths.js");
const serialize_js_1 = require("../utils/serialize.js");
const branch_js_1 = require("./branch.js");
const compose_js_1 = require("./compose.js");
const props_js_1 = require("./rules/props.js");
const flattenBranches = (branches, ctx) => {
    const discriminants = calculateDiscriminants(branches, ctx);
    const indices = branches.map((_, i) => i);
    return discriminate(branches, indices, discriminants, ctx);
};
exports.flattenBranches = flattenBranches;
const discriminate = (originalBranches, remainingIndices, discriminants, ctx) => {
    if (remainingIndices.length === 1) {
        return (0, branch_js_1.flattenBranch)(originalBranches[remainingIndices[0]], ctx);
    }
    const bestDiscriminant = findBestDiscriminant(remainingIndices, discriminants);
    if (!bestDiscriminant) {
        return [
            [
                "branches",
                remainingIndices.map((i) => branchIncludesMorph(originalBranches[i], ctx.type.scope)
                    ? (0, errors_js_1.throwParseError)((0, union_js_1.writeUndiscriminatableMorphUnionMessage)(`${ctx.path}`))
                    : (0, branch_js_1.flattenBranch)(originalBranches[i], ctx))
            ]
        ];
    }
    const cases = {};
    for (const caseKey in bestDiscriminant.indexCases) {
        const nextIndices = bestDiscriminant.indexCases[caseKey];
        cases[caseKey] = discriminate(originalBranches, nextIndices, discriminants, ctx);
        if (caseKey !== "default") {
            pruneDiscriminant(cases[caseKey], bestDiscriminant.path, bestDiscriminant, ctx);
        }
    }
    return [
        [
            "switch",
            {
                path: bestDiscriminant.path,
                kind: bestDiscriminant.kind,
                cases
            }
        ]
    ];
};
const pruneDiscriminant = (entries, segments, discriminant, ctx) => {
    for (let i = 0; i < entries.length; i++) {
        const [k, v] = entries[i];
        if (!segments.length) {
            if (discriminant.kind === "domain") {
                if (k === "domain" || k === "domains") {
                    entries.splice(i, 1);
                    return;
                }
                else if (k === "class" || k === "value") {
                    // these keys imply a domain, but also add additional
                    // information, so can't be pruned
                    return;
                }
            }
            else if (discriminant.kind === k) {
                entries.splice(i, 1);
                return;
            }
        }
        else if ((k === "requiredProp" ||
            k === "prerequisiteProp" ||
            k === "optionalProp") &&
            v[0] === segments[0]) {
            if (typeof v[1] === "string") {
                if (discriminant.kind !== "domain") {
                    return throwInternalPruneFailure(discriminant);
                }
                entries.splice(i, 1);
                return;
            }
            pruneDiscriminant(v[1], segments.slice(1), discriminant, ctx);
            if (v[1].length === 0) {
                entries.splice(i, 1);
            }
            return;
        }
        // check for branch keys, which must be traversed even if there are no
        // segments left
        if (k === "domains") {
            /* c8 ignore next */
            if ((0, generics_js_1.keyCount)(v) !== 1 || !v.object) {
                return throwInternalPruneFailure(discriminant);
            }
            pruneDiscriminant(v.object, segments, discriminant, ctx);
            return;
        }
        else if (k === "switch") {
            for (const caseKey in v.cases) {
                pruneDiscriminant(v.cases[caseKey], segments, discriminant, ctx);
            }
            return;
        }
        else if (k === "branches") {
            for (const branch of v) {
                pruneDiscriminant(branch, segments, discriminant, ctx);
            }
            return;
        }
    }
    return throwInternalPruneFailure(discriminant);
};
const throwInternalPruneFailure = (discriminant) => (0, errors_js_1.throwInternalError)(`Unexpectedly failed to discriminate ${discriminant.kind} at path '${discriminant.path}'`);
const discriminantKinds = {
    domain: true,
    class: true,
    value: true
};
const calculateDiscriminants = (branches, ctx) => {
    const discriminants = {
        disjointsByPair: {},
        casesByDisjoint: {}
    };
    for (let lIndex = 0; lIndex < branches.length - 1; lIndex++) {
        for (let rIndex = lIndex + 1; rIndex < branches.length; rIndex++) {
            const pairKey = `${lIndex}/${rIndex}`;
            const pairDisjoints = [];
            discriminants.disjointsByPair[pairKey] = pairDisjoints;
            const intersectionState = new compose_js_1.IntersectionState(ctx.type, "|");
            (0, branch_js_1.branchIntersection)(branches[lIndex], branches[rIndex], intersectionState);
            for (const path in intersectionState.disjoints) {
                if (path.includes(props_js_1.mappedKeys.index)) {
                    // containers could be empty and therefore their elements cannot be used to discriminate
                    // allowing this via a special case where both are length >0 tracked here:
                    // https://github.com/arktypeio/arktype/issues/593
                    continue;
                }
                const { l, r, kind } = intersectionState.disjoints[path];
                if (!(0, generics_js_1.isKeyOf)(kind, discriminantKinds)) {
                    continue;
                }
                const lSerialized = (0, exports.serializeDefinitionIfAllowed)(kind, l);
                const rSerialized = (0, exports.serializeDefinitionIfAllowed)(kind, r);
                if (lSerialized === undefined || rSerialized === undefined) {
                    continue;
                }
                const qualifiedDisjoint = path === "/" ? kind : `${path}/${kind}`;
                pairDisjoints.push(qualifiedDisjoint);
                if (!discriminants.casesByDisjoint[qualifiedDisjoint]) {
                    discriminants.casesByDisjoint[qualifiedDisjoint] = {
                        [lSerialized]: [lIndex],
                        [rSerialized]: [rIndex]
                    };
                    continue;
                }
                const cases = discriminants.casesByDisjoint[qualifiedDisjoint];
                const existingLBranch = cases[lSerialized];
                if (!existingLBranch) {
                    cases[lSerialized] = [lIndex];
                }
                else if (!existingLBranch.includes(lIndex)) {
                    existingLBranch.push(lIndex);
                }
                const existingRBranch = cases[rSerialized];
                if (!existingRBranch) {
                    cases[rSerialized] = [rIndex];
                }
                else if (!existingRBranch.includes(rIndex)) {
                    existingRBranch.push(rIndex);
                }
            }
        }
    }
    return discriminants;
};
const parseQualifiedDisjoint = (qualifiedDisjoint) => {
    const path = paths_js_1.Path.fromString(qualifiedDisjoint);
    return [path, path.pop()];
};
const findBestDiscriminant = (remainingIndices, discriminants) => {
    let bestDiscriminant;
    for (let i = 0; i < remainingIndices.length - 1; i++) {
        const lIndex = remainingIndices[i];
        for (let j = i + 1; j < remainingIndices.length; j++) {
            const rIndex = remainingIndices[j];
            const candidates = discriminants.disjointsByPair[`${lIndex}/${rIndex}`];
            for (const qualifiedDisjoint of candidates) {
                const indexCases = discriminants.casesByDisjoint[qualifiedDisjoint];
                const filteredCases = {};
                const defaultCases = [
                    ...remainingIndices
                ];
                let score = 0;
                for (const caseKey in indexCases) {
                    const filteredIndices = indexCases[caseKey].filter((i) => {
                        const remainingIndex = remainingIndices.indexOf(i);
                        if (remainingIndex !== -1) {
                            delete defaultCases[remainingIndex];
                            return true;
                        }
                    });
                    if (filteredIndices.length === 0) {
                        continue;
                    }
                    filteredCases[caseKey] = filteredIndices;
                    score++;
                }
                const defaultCaseKeys = (0, generics_js_1.objectKeysOf)(defaultCases);
                if (defaultCaseKeys.length) {
                    filteredCases["default"] = defaultCaseKeys.map((k) => parseInt(k));
                }
                if (!bestDiscriminant || score > bestDiscriminant.score) {
                    const [path, kind] = parseQualifiedDisjoint(qualifiedDisjoint);
                    bestDiscriminant = {
                        path,
                        kind,
                        indexCases: filteredCases,
                        score
                    };
                    if (score === remainingIndices.length) {
                        // if we find a candidate that discriminates all branches, return early
                        return bestDiscriminant;
                    }
                }
            }
        }
    }
    return bestDiscriminant;
};
const serializeDefinitionIfAllowed = (kind, definition) => {
    switch (kind) {
        case "value":
            return serializeIfPrimitive(definition);
        case "domain":
            return definition;
        case "class":
            return (0, objectKinds_js_1.getExactConstructorObjectKind)(definition);
        default:
            return;
    }
};
exports.serializeDefinitionIfAllowed = serializeDefinitionIfAllowed;
const serializeIfPrimitive = (data) => {
    const domain = (0, domains_js_1.domainOf)(data);
    return domain === "object" || domain === "symbol"
        ? undefined
        : (0, serialize_js_1.serializePrimitive)(data);
};
const serializeData = {
    value: (data) => serializeIfPrimitive(data) ?? "default",
    class: (data) => (0, objectKinds_js_1.objectKindOf)(data) ?? "default",
    domain: domains_js_1.domainOf
};
const serializeCase = (kind, data) => serializeData[kind](data);
exports.serializeCase = serializeCase;
const branchIncludesMorph = (branch, $) => "morph" in branch
    ? true
    : "props" in branch
        ? Object.values(branch.props).some((prop) => nodeIncludesMorph((0, props_js_1.propToNode)(prop), $))
        : false;
const nodeIncludesMorph = (node, $) => typeof node === "string"
    ? $.resolve(node).includesMorph
    : Object.values($.resolveTypeNode(node)).some((predicate) => predicate === true
        ? false
        : (0, objectKinds_js_1.isArray)(predicate)
            ? predicate.some((branch) => branchIncludesMorph(branch, $))
            : branchIncludesMorph(predicate, $));
