"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.flattenProps = exports.propsIntersection = exports.propToNode = exports.mappedKeys = exports.isPrerequisite = exports.isOptional = void 0;
const compose_js_1 = require("../compose.js");
const node_js_1 = require("../node.js");
const isOptional = (prop) => prop[0] === "?";
exports.isOptional = isOptional;
const isPrerequisite = (prop) => prop[0] === "!";
exports.isPrerequisite = isPrerequisite;
exports.mappedKeys = {
    index: "[index]"
};
const propToNode = (prop) => (0, exports.isOptional)(prop) || (0, exports.isPrerequisite)(prop) ? prop[1] : prop;
exports.propToNode = propToNode;
const getTupleLengthIfPresent = (result) => {
    if (typeof result.length === "object" &&
        (0, exports.isPrerequisite)(result.length) &&
        typeof result.length[1] !== "string" &&
        (0, node_js_1.isLiteralNode)(result.length[1], "number")) {
        return result.length[1].number.value;
    }
};
exports.propsIntersection = (0, compose_js_1.composeIntersection)((l, r, state) => {
    const result = propKeysIntersection(l, r, state);
    if (typeof result === "symbol") {
        return result;
    }
    const lengthValue = getTupleLengthIfPresent(result);
    if (lengthValue === undefined || !(exports.mappedKeys.index in result)) {
        return result;
    }
    // if we are at this point, we have an array with an exact length (i.e.
    // a tuple) and an index signature. Intersection each tuple item with
    // the index signature node and remove the index signature via a new
    // updated result, copied from result to avoid mutating existing references.
    const { [exports.mappedKeys.index]: indexProp, ...updatedResult } = result;
    const indexNode = (0, exports.propToNode)(indexProp);
    for (let i = 0; i < lengthValue; i++) {
        if (!updatedResult[i]) {
            updatedResult[i] = indexNode;
            continue;
        }
        const existingNodeAtIndex = (0, exports.propToNode)(updatedResult[i]);
        state.path.push(`${i}`);
        const updatedResultAtIndex = (0, node_js_1.nodeIntersection)(existingNodeAtIndex, indexNode, state);
        state.path.pop();
        if ((0, compose_js_1.isDisjoint)(updatedResultAtIndex)) {
            return updatedResultAtIndex;
        }
        else if (!(0, compose_js_1.isEquality)(updatedResultAtIndex) &&
            updatedResultAtIndex !== existingNodeAtIndex) {
            updatedResult[i] = updatedResultAtIndex;
        }
    }
    return updatedResult;
});
const propKeysIntersection = (0, compose_js_1.composeKeyedIntersection)((propKey, l, r, context) => {
    if (l === undefined) {
        return r === undefined ? (0, compose_js_1.equality)() : r;
    }
    if (r === undefined) {
        return l;
    }
    context.path.push(propKey);
    const result = (0, node_js_1.nodeIntersection)((0, exports.propToNode)(l), (0, exports.propToNode)(r), context);
    context.path.pop();
    const resultIsOptional = (0, exports.isOptional)(l) && (0, exports.isOptional)(r);
    if ((0, compose_js_1.isDisjoint)(result) && resultIsOptional) {
        // If an optional key has an empty intersection, the type can
        // still be satisfied as long as the key is not included. Set
        // the node to never rather than invalidating the type.
        return {};
    }
    return result;
}, { onEmpty: "bubble" });
const flattenProps = (entries, props, ctx) => {
    const keyConfig = ctx.type.config?.keys ?? ctx.type.scope.config.keys;
    return keyConfig === "loose"
        ? flattenLooseProps(entries, props, ctx)
        : flattenPropsRecord(keyConfig, entries, props, ctx);
};
exports.flattenProps = flattenProps;
const flattenLooseProps = (entries, props, ctx) => {
    // if we don't care about extraneous keys, flatten props so we can iterate over the definitions directly
    for (const k in props) {
        const prop = props[k];
        ctx.path.push(k);
        if (k === exports.mappedKeys.index) {
            entries.push(["indexProp", (0, node_js_1.flattenNode)((0, exports.propToNode)(prop), ctx)]);
        }
        else if ((0, exports.isOptional)(prop)) {
            entries.push(["optionalProp", [k, (0, node_js_1.flattenNode)(prop[1], ctx)]]);
        }
        else if ((0, exports.isPrerequisite)(prop)) {
            entries.push(["prerequisiteProp", [k, (0, node_js_1.flattenNode)(prop[1], ctx)]]);
        }
        else {
            entries.push(["requiredProp", [k, (0, node_js_1.flattenNode)(prop, ctx)]]);
        }
        ctx.path.pop();
    }
};
const flattenPropsRecord = (kind, entries, props, ctx) => {
    const result = {
        required: {},
        optional: {}
    };
    // if we need to keep track of extraneous keys, either to add problems or
    // remove them, store the props as a Record to optimize for presence
    // checking as we iterate over the data
    for (const k in props) {
        const prop = props[k];
        ctx.path.push(k);
        if (k === exports.mappedKeys.index) {
            result.index = (0, node_js_1.flattenNode)((0, exports.propToNode)(prop), ctx);
        }
        else if ((0, exports.isOptional)(prop)) {
            result.optional[k] = (0, node_js_1.flattenNode)(prop[1], ctx);
        }
        else if ((0, exports.isPrerequisite)(prop)) {
            // we still have to push prerequisite props as normal entries so they can be checked first
            entries.push(["prerequisiteProp", [k, (0, node_js_1.flattenNode)(prop[1], ctx)]]);
        }
        else {
            result.required[k] = (0, node_js_1.flattenNode)(prop, ctx);
        }
        ctx.path.pop();
    }
    entries.push([`${kind}Props`, result]);
};
