"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseKeyOfTuple = void 0;
const props_js_1 = require("../../nodes/rules/props.js");
const errors_js_1 = require("../../utils/errors.js");
const freeze_js_1 = require("../../utils/freeze.js");
const generics_js_1 = require("../../utils/generics.js");
const numericLiterals_js_1 = require("../../utils/numericLiterals.js");
const objectKinds_js_1 = require("../../utils/objectKinds.js");
const serialize_js_1 = require("../../utils/serialize.js");
const definition_js_1 = require("../definition.js");
const intersection_js_1 = require("./intersection.js");
const arrayIndexStringBranch = (0, freeze_js_1.deepFreeze)({
    regex: numericLiterals_js_1.wellFormedNonNegativeIntegerMatcher.source
});
const arrayIndexNumberBranch = (0, freeze_js_1.deepFreeze)({
    range: { min: { comparator: ">=", limit: 0 } },
    divisor: 1
});
const parseKeyOfTuple = (def, ctx) => {
    const resolution = ctx.type.scope.resolveNode((0, definition_js_1.parseDefinition)(def[1], ctx));
    const predicateKeys = (0, generics_js_1.objectKeysOf)(resolution).map((domain) => keysOfPredicate(domain, resolution[domain]));
    const sharedKeys = sharedKeysOf(predicateKeys);
    if (!sharedKeys.length) {
        return (0, intersection_js_1.writeImplicitNeverMessage)(ctx.path, "keyof");
    }
    const keyNode = {};
    for (const key of sharedKeys) {
        const keyType = typeof key;
        if (keyType === "string" ||
            keyType === "number" ||
            keyType === "symbol") {
            keyNode[keyType] ??= [];
            keyNode[keyType].push({ value: key });
        }
        else if (key === numericLiterals_js_1.wellFormedNonNegativeIntegerMatcher) {
            keyNode.string ??= [];
            keyNode.string.push(arrayIndexStringBranch);
            keyNode.number ??= [];
            keyNode.number.push(arrayIndexNumberBranch);
        }
        else {
            return (0, errors_js_1.throwInternalError)(`Unexpected keyof key '${(0, serialize_js_1.stringify)(key)}'`);
        }
    }
    return Object.fromEntries(Object.entries(keyNode).map(([domain, branches]) => [
        domain,
        branches.length === 1 ? branches[0] : branches
    ]));
};
exports.parseKeyOfTuple = parseKeyOfTuple;
const baseKeysByDomain = {
    bigint: (0, generics_js_1.prototypeKeysOf)(0n),
    boolean: (0, generics_js_1.prototypeKeysOf)(false),
    null: [],
    number: (0, generics_js_1.prototypeKeysOf)(0),
    // TS doesn't include the Object prototype in keyof, so keyof object is never
    object: [],
    string: (0, generics_js_1.prototypeKeysOf)(""),
    symbol: (0, generics_js_1.prototypeKeysOf)(Symbol()),
    undefined: []
};
const keysOfPredicate = (domain, predicate) => domain !== "object" || predicate === true
    ? baseKeysByDomain[domain]
    : sharedKeysOf((0, generics_js_1.listFrom)(predicate).map((branch) => keysOfObjectBranch(branch)));
const sharedKeysOf = (keyBranches) => {
    if (!keyBranches.length) {
        return [];
    }
    let sharedKeys = keyBranches[0];
    for (let i = 1; i < keyBranches.length; i++) {
        // we can filter directly by equality here because the RegExp we're
        // using will always be reference equal to
        // wellFormedNonNegativeIntegerMatcher
        sharedKeys = sharedKeys.filter((k) => keyBranches[i].includes(k));
    }
    return sharedKeys;
};
const keysOfObjectBranch = (branch) => {
    const result = [];
    if ("props" in branch) {
        for (const key of Object.keys(branch.props)) {
            if (key === props_js_1.mappedKeys.index) {
                // if any number is a valid key push this RegExp
                result.push(numericLiterals_js_1.wellFormedNonNegativeIntegerMatcher);
            }
            else if (!result.includes(key)) {
                result.push(key);
                if (numericLiterals_js_1.wellFormedNonNegativeIntegerMatcher.test(key)) {
                    // allow numeric access to keys
                    result.push((0, numericLiterals_js_1.tryParseWellFormedInteger)(key, `Unexpectedly failed to parse an integer from key '${key}'`));
                }
            }
        }
    }
    if ("class" in branch) {
        const constructor = typeof branch.class === "string"
            ? objectKinds_js_1.defaultObjectKinds[branch.class]
            : branch.class;
        for (const key of (0, generics_js_1.prototypeKeysOf)(constructor.prototype)) {
            if (!result.includes(key)) {
                result.push(key);
            }
        }
    }
    return result;
};
