"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DynamicState = void 0;
const node_js_1 = require("../../../nodes/node.js");
const range_js_1 = require("../../../nodes/rules/range.js");
const errors_js_1 = require("../../../utils/errors.js");
const generics_js_1 = require("../../../utils/generics.js");
const serialize_js_1 = require("../../../utils/serialize.js");
const scanner_js_1 = require("../shift/scanner.js");
const shared_js_1 = require("./shared.js");
class DynamicState {
    ctx;
    scanner;
    root;
    branches = {};
    groups = [];
    constructor(def, ctx) {
        this.ctx = ctx;
        this.scanner = new scanner_js_1.Scanner(def);
    }
    error(message) {
        return (0, errors_js_1.throwParseError)(message);
    }
    hasRoot() {
        return this.root !== undefined;
    }
    resolveRoot() {
        this.assertHasRoot();
        return this.ctx.type.scope.resolveTypeNode(this.root);
    }
    rootToString() {
        this.assertHasRoot();
        return (0, serialize_js_1.stringify)(this.root);
    }
    ejectRootIfLimit() {
        this.assertHasRoot();
        const resolution = typeof this.root === "string"
            ? this.ctx.type.scope.resolveNode(this.root)
            : this.root;
        if ((0, node_js_1.isLiteralNode)(resolution, "number")) {
            const limit = resolution.number.value;
            this.root = undefined;
            return limit;
        }
    }
    ejectRangeIfOpen() {
        if (this.branches.range) {
            const range = this.branches.range;
            delete this.branches.range;
            return range;
        }
    }
    assertHasRoot() {
        if (this.root === undefined) {
            return (0, errors_js_1.throwInternalError)("Unexpected interaction with unset root");
        }
    }
    assertUnsetRoot() {
        if (this.root !== undefined) {
            return (0, errors_js_1.throwInternalError)("Unexpected attempt to overwrite root");
        }
    }
    setRoot(node) {
        this.assertUnsetRoot();
        this.root = node;
    }
    rootToArray() {
        this.root = (0, node_js_1.toArrayNode)(this.ejectRoot());
    }
    intersect(node) {
        this.root = (0, node_js_1.rootIntersection)(this.ejectRoot(), node, this.ctx.type);
    }
    ejectRoot() {
        this.assertHasRoot();
        const root = this.root;
        this.root = undefined;
        return root;
    }
    ejectFinalizedRoot() {
        this.assertHasRoot();
        const root = this.root;
        this.root = ejectedProxy;
        return root;
    }
    finalize() {
        if (this.groups.length) {
            return this.error(shared_js_1.unclosedGroupMessage);
        }
        this.finalizeBranches();
        this.scanner.finalized = true;
    }
    reduceLeftBound(limit, comparator) {
        const invertedComparator = scanner_js_1.Scanner.invertedComparators[comparator];
        if (!(0, generics_js_1.isKeyOf)(invertedComparator, range_js_1.minComparators)) {
            return this.error((0, shared_js_1.writeUnpairableComparatorMessage)(comparator));
        }
        if (this.branches.range) {
            return this.error((0, shared_js_1.writeMultipleLeftBoundsMessage)(`${this.branches.range.limit}`, this.branches.range.comparator, `${limit}`, invertedComparator));
        }
        this.branches.range = {
            limit,
            comparator: invertedComparator
        };
    }
    finalizeBranches() {
        this.assertRangeUnset();
        if (this.branches.union) {
            this.pushRootToBranch("|");
            this.setRoot(this.branches.union);
        }
        else if (this.branches.intersection) {
            this.setRoot((0, node_js_1.rootIntersection)(this.branches.intersection, this.ejectRoot(), this.ctx.type));
        }
    }
    finalizeGroup() {
        this.finalizeBranches();
        const topBranchState = this.groups.pop();
        if (!topBranchState) {
            return this.error((0, shared_js_1.writeUnmatchedGroupCloseMessage)(this.scanner.unscanned));
        }
        this.branches = topBranchState;
    }
    pushRootToBranch(token) {
        this.assertRangeUnset();
        this.branches.intersection = this.branches.intersection
            ? (0, node_js_1.rootIntersection)(this.branches.intersection, this.ejectRoot(), this.ctx.type)
            : this.ejectRoot();
        if (token === "|") {
            this.branches.union = this.branches.union
                ? (0, node_js_1.rootUnion)(this.branches.union, this.branches.intersection, this.ctx.type)
                : this.branches.intersection;
            delete this.branches.intersection;
        }
    }
    assertRangeUnset() {
        if (this.branches.range) {
            return this.error((0, shared_js_1.writeOpenRangeMessage)(`${this.branches.range.limit}`, this.branches.range.comparator));
        }
    }
    reduceGroupOpen() {
        this.groups.push(this.branches);
        this.branches = {};
    }
    previousOperator() {
        return this.branches.range?.comparator ?? this.branches.intersection
            ? "&"
            : this.branches.union
                ? "|"
                : undefined;
    }
    shiftedByOne() {
        this.scanner.shift();
        return this;
    }
}
exports.DynamicState = DynamicState;
const ejectedProxy = new Proxy({}, {
    get: () => (0, errors_js_1.throwInternalError)(`Unexpected attempt to access ejected attributes`)
});
