"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.writeEmptyRangeMessage = exports.writeInvalidLimitMessage = exports.parseRightBound = exports.singleEqualsMessage = exports.parseBound = void 0;
const compose_js_1 = require("../../../../nodes/compose.js");
const range_js_1 = require("../../../../nodes/rules/range.js");
const errors_js_1 = require("../../../../utils/errors.js");
const generics_js_1 = require("../../../../utils/generics.js");
const numericLiterals_js_1 = require("../../../../utils/numericLiterals.js");
const bound_js_1 = require("../../../ast/bound.js");
const shared_js_1 = require("../../reduce/shared.js");
const scanner_js_1 = require("../scanner.js");
const parseBound = (s, start) => {
    const comparator = shiftComparator(s, start);
    const maybeMin = s.ejectRootIfLimit();
    return maybeMin === undefined
        ? (0, exports.parseRightBound)(s, comparator)
        : s.reduceLeftBound(maybeMin, comparator);
};
exports.parseBound = parseBound;
const shiftComparator = (s, start) => s.scanner.lookaheadIs("=")
    ? `${start}${s.scanner.shift()}`
    : (0, generics_js_1.isKeyOf)(start, scanner_js_1.Scanner.oneCharComparators)
        ? start
        : s.error(exports.singleEqualsMessage);
exports.singleEqualsMessage = `= is not a valid comparator. Use == to check for equality`;
const parseRightBound = (s, comparator) => {
    const limitToken = s.scanner.shiftUntilNextTerminator();
    const limit = (0, numericLiterals_js_1.tryParseWellFormedNumber)(limitToken, (0, exports.writeInvalidLimitMessage)(comparator, limitToken + s.scanner.unscanned));
    const openRange = s.ejectRangeIfOpen();
    const rightBound = { comparator, limit };
    const range = openRange
        ? !hasComparatorIn(rightBound, range_js_1.maxComparators)
            ? s.error((0, shared_js_1.writeUnpairableComparatorMessage)(comparator))
            : (0, range_js_1.compareStrictness)("min", openRange, rightBound) === "l"
                ? s.error((0, exports.writeEmptyRangeMessage)({ min: openRange, max: rightBound }))
                : {
                    min: openRange,
                    max: rightBound
                }
        : hasComparator(rightBound, "==")
            ? rightBound
            : hasComparatorIn(rightBound, range_js_1.minComparators)
                ? { min: rightBound }
                : hasComparatorIn(rightBound, range_js_1.maxComparators)
                    ? { max: rightBound }
                    : (0, errors_js_1.throwInternalError)(`Unexpected comparator '${rightBound.comparator}'`);
    s.intersect(distributeRange(range, s));
};
exports.parseRightBound = parseRightBound;
const distributeRange = (range, s) => {
    const resolution = s.resolveRoot();
    const domains = (0, generics_js_1.objectKeysOf)(resolution);
    const distributedRange = {};
    const rangePredicate = { range };
    const isBoundable = domains.every((domain) => {
        switch (domain) {
            case "string":
                distributedRange.string = rangePredicate;
                return true;
            case "number":
                distributedRange.number = rangePredicate;
                return true;
            case "object":
                distributedRange.object = rangePredicate;
                if (resolution.object === true) {
                    return false;
                }
                return (0, generics_js_1.listFrom)(resolution.object).every((branch) => "class" in branch && branch.class === Array);
            default:
                return false;
        }
    });
    if (!isBoundable) {
        s.error((0, bound_js_1.writeUnboundableMessage)(s.rootToString()));
    }
    return distributedRange;
};
const hasComparator = (bound, comparator) => bound.comparator === comparator;
const hasComparatorIn = (bound, comparators) => bound.comparator in comparators;
const writeInvalidLimitMessage = (comparator, limit) => `Comparator ${comparator} must be followed by a number literal (was '${limit}')`;
exports.writeInvalidLimitMessage = writeInvalidLimitMessage;
const writeEmptyRangeMessage = (range) => `${(0, compose_js_1.stringifyRange)(range)} is empty`;
exports.writeEmptyRangeMessage = writeEmptyRangeMessage;
