"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.writeDuplicateAliasesMessage = exports.writeShallowCycleErrorMessage = exports.rootType = exports.rootScope = exports.scope = exports.Scope = exports.isConfigTuple = exports.compileScopeOptions = void 0;
const node_js_1 = require("../nodes/node.js");
const definition_js_1 = require("../parse/definition.js");
const problems_js_1 = require("../traverse/problems.js");
const chainableNoOpProxy_js_1 = require("../utils/chainableNoOpProxy.js");
const errors_js_1 = require("../utils/errors.js");
const freeze_js_1 = require("../utils/freeze.js");
const generics_js_1 = require("../utils/generics.js");
const paths_js_1 = require("../utils/paths.js");
const cache_js_1 = require("./cache.js");
const type_js_1 = require("./type.js");
const compileScopeOptions = (opts) => ({
    codes: (0, problems_js_1.compileProblemWriters)(opts.codes),
    keys: opts.keys ?? "loose"
});
exports.compileScopeOptions = compileScopeOptions;
let anonymousScopeCount = 0;
const scopeRegistry = {};
const spaceRegistry = {};
const isConfigTuple = (def) => Array.isArray(def) && def[1] === ":";
exports.isConfigTuple = isConfigTuple;
class Scope {
    aliases;
    name;
    config;
    parseCache = new cache_js_1.FreezingCache();
    #resolutions = new cache_js_1.Cache();
    #exports = new cache_js_1.Cache();
    constructor(aliases, opts = {}) {
        this.aliases = aliases;
        this.name = this.#register(opts);
        if (opts.standard !== false) {
            this.#cacheSpaces([spaceRegistry["standard"]], "imports");
        }
        if (opts.imports) {
            this.#cacheSpaces(opts.imports, "imports");
        }
        if (opts.includes) {
            this.#cacheSpaces(opts.includes, "includes");
        }
        this.config = (0, exports.compileScopeOptions)(opts);
    }
    #register(opts) {
        const name = opts.name
            ? scopeRegistry[opts.name]
                ? (0, errors_js_1.throwParseError)(`A scope named '${opts.name}' already exists`)
                : opts.name
            : `scope${++anonymousScopeCount}`;
        scopeRegistry[name] = this;
        return name;
    }
    #cacheSpaces(spaces, kind) {
        for (const space of spaces) {
            for (const name in space) {
                if (this.#resolutions.has(name) || name in this.aliases) {
                    (0, errors_js_1.throwParseError)((0, exports.writeDuplicateAliasesMessage)(name));
                }
                this.#resolutions.set(name, space[name]);
                if (kind === "includes") {
                    this.#exports.set(name, space[name]);
                }
            }
        }
    }
    #initializeContext(type) {
        return {
            type,
            path: new paths_js_1.Path()
        };
    }
    getAnonymousQualifiedName(base) {
        let increment = 0;
        let id = base;
        while (this.isResolvable(id)) {
            id = `${base}${increment++}`;
        }
        return `${this.name}.${id}`;
    }
    addAnonymousTypeReference(referencedType, ctx) {
        ctx.type.includesMorph ||= referencedType.includesMorph;
        return referencedType.node;
    }
    get infer() {
        return chainableNoOpProxy_js_1.chainableNoOpProxy;
    }
    compile() {
        if (!spaceRegistry[this.name]) {
            for (const name in this.aliases) {
                this.resolve(name);
            }
            spaceRegistry[this.name] = this.#exports.root;
        }
        return this.#exports.root;
    }
    addParsedReferenceIfResolvable(name, ctx) {
        const resolution = this.#resolveRecurse(name, "undefined", [name]);
        if (!resolution) {
            return false;
        }
        ctx.type.includesMorph ||= resolution.includesMorph;
        return true;
    }
    resolve(name) {
        return this.#resolveRecurse(name, "throw", [name]);
    }
    #resolveRecurse(name, onUnresolvable, seen) {
        const maybeCacheResult = this.#resolutions.get(name);
        if (maybeCacheResult) {
            return maybeCacheResult;
        }
        const aliasDef = this.aliases[name];
        if (!aliasDef) {
            return (onUnresolvable === "throw"
                ? (0, errors_js_1.throwInternalError)(`Unexpectedly failed to resolve alias '${name}'`)
                : undefined);
        }
        const t = (0, type_js_1.initializeType)(name, aliasDef, {}, this);
        const ctx = this.#initializeContext(t);
        this.#resolutions.set(name, t);
        this.#exports.set(name, t);
        let node = (0, definition_js_1.parseDefinition)(aliasDef, ctx);
        if (typeof node === "string") {
            if (seen.includes(node)) {
                return (0, errors_js_1.throwParseError)((0, exports.writeShallowCycleErrorMessage)(name, seen));
            }
            seen.push(node);
            node = this.#resolveRecurse(node, "throw", seen).node;
        }
        t.node = (0, freeze_js_1.deepFreeze)(node);
        t.flat = (0, freeze_js_1.deepFreeze)((0, node_js_1.flattenType)(t));
        return t;
    }
    resolveNode(node) {
        return typeof node === "string"
            ? this.resolveNode(this.resolve(node).node)
            : node;
    }
    resolveTypeNode(node) {
        const resolution = this.resolveNode(node);
        return (0, node_js_1.isConfigNode)(resolution) ? resolution.node : resolution;
    }
    expressions = {
        intersection: (l, r, opts) => this.type([l, "&", r], opts),
        union: (l, r, opts) => this.type([l, "|", r], opts),
        arrayOf: (def, opts) => this.type([def, "[]"], opts),
        keyOf: (def, opts) => this.type(["keyof", def], opts),
        node: (def, opts) => this.type(["node", def], opts),
        instanceOf: (def, opts) => this.type(["instanceof", def], opts),
        valueOf: (def, opts) => this.type(["===", def], opts),
        narrow: (def, fn, opts) => this.type([def, "=>", fn], opts),
        morph: (def, fn, opts) => this.type([def, "|>", fn], opts)
    };
    intersection = this.expressions.intersection;
    union = this.expressions.union;
    arrayOf = this.expressions.arrayOf;
    keyOf = this.expressions.keyOf;
    valueOf = this.expressions.valueOf;
    instanceOf = this.expressions.instanceOf;
    narrow = this.expressions.narrow;
    morph = this.expressions.morph;
    type = Object.assign((def, config = {}) => {
        const t = (0, type_js_1.initializeType)("λtype", def, config, this);
        const ctx = this.#initializeContext(t);
        const root = (0, definition_js_1.parseDefinition)(def, ctx);
        t.node = (0, freeze_js_1.deepFreeze)((0, generics_js_1.hasKeys)(config)
            ? { config, node: this.resolveTypeNode(root) }
            : root);
        t.flat = (0, freeze_js_1.deepFreeze)((0, node_js_1.flattenType)(t));
        return t;
    }, { from: this.expressions.node });
    isResolvable(name) {
        return this.#resolutions.has(name) || this.aliases[name];
    }
}
exports.Scope = Scope;
exports.scope = ((aliases, opts = {}) => new Scope(aliases, opts));
exports.rootScope = (0, exports.scope)({}, { name: "root", standard: false });
exports.rootType = exports.rootScope.type;
const writeShallowCycleErrorMessage = (name, seen) => `Alias '${name}' has a shallow resolution cycle: ${[...seen, name].join("=>")}`;
exports.writeShallowCycleErrorMessage = writeShallowCycleErrorMessage;
const writeDuplicateAliasesMessage = (name) => `Alias '${name}' is already defined`;
exports.writeDuplicateAliasesMessage = writeDuplicateAliasesMessage;
