"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.validation = exports.validationScope = void 0;
const numericLiterals_js_1 = require("../../utils/numericLiterals.js");
const scope_js_1 = require("../scope.js");
const tsKeywords_js_1 = require("../tsKeywords.js");
const creditCard_js_1 = require("./creditCard.js");
const date_js_1 = require("./date.js");
// Non-trivial expressions should have an explanation or attribution
const parsedNumber = (0, scope_js_1.rootType)([numericLiterals_js_1.wellFormedNumberMatcher, "|>", (s) => parseFloat(s)], { mustBe: "a well-formed numeric string" });
const parsedInteger = (0, scope_js_1.rootType)([
    tsKeywords_js_1.tsKeywords.string,
    "|>",
    (s, problems) => {
        if (!(0, numericLiterals_js_1.isWellFormedInteger)(s)) {
            return problems.mustBe("a well-formed integer string");
        }
        const parsed = parseInt(s);
        return Number.isSafeInteger(parsed)
            ? parsed
            : problems.mustBe("an integer in the range Number.MIN_SAFE_INTEGER to Number.MAX_SAFE_INTEGER");
    }
]);
// https://www.regular-expressions.info/email.html
const email = (0, scope_js_1.rootType)(/^[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Za-z]{2,}$/, {
    mustBe: "a valid email"
});
// https://github.com/validatorjs/validator.js/blob/master/src/lib/isUUID.js
const uuid = (0, scope_js_1.rootType)(/^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-[0-9A-Fa-f]{4}-[0-9A-Fa-f]{4}-[0-9A-Fa-f]{12}$/, { mustBe: "a valid UUID" });
// https://semver.org/
const semver = (0, scope_js_1.rootType)(/^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(?:-((?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?(?:\+([0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?$/, { mustBe: "a valid semantic version (see https://semver.org/)" });
const json = (0, scope_js_1.rootType)([tsKeywords_js_1.tsKeywords.string, "|>", (s) => JSON.parse(s)], {
    mustBe: "a JSON-parsable string"
});
/**
 * @keywords keywords: {
        "alpha": "only letters",
        "alphanumeric": "only letters and digits",
        "lowercase": "only lowercase letters",
        "uppercase": "only uppercase letters",
        "creditCard": "a valid credit card number",
        "email": "a valid email",
        "uuid": "a valid UUID",
        "parsedNumber": "a well-formed numeric string",
        "parsedInteger": "a well-formed integer string",
        "parsedDate": "a valid date",
        "semver": "a valid semantic version",
        "json": "a JSON-parsable string",
        "integer": "an integer"
}
 * @docgenScope
 * @docgenTable
 */
exports.validationScope = (0, scope_js_1.scope)({
    // Character sets
    alpha: [/^[A-Za-z]*$/, ":", { mustBe: "only letters" }],
    alphanumeric: [
        /^[A-Za-z\d]*$/,
        ":",
        { mustBe: "only letters and digits" }
    ],
    lowercase: [/^[a-z]*$/, ":", { mustBe: "only lowercase letters" }],
    uppercase: [/^[A-Z]*$/, ":", { mustBe: "only uppercase letters" }],
    creditCard: creditCard_js_1.creditCard,
    email,
    uuid,
    parsedNumber,
    parsedInteger,
    parsedDate: date_js_1.parsedDate,
    semver,
    json,
    integer: ["node", { number: { divisor: 1 } }]
}, { name: "validation", standard: false });
exports.validation = exports.validationScope.compile();
