"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.compileProblemWriters = exports.defaultProblemWriters = exports.problemCodes = exports.describeBranches = exports.objectKindsToDescriptions = exports.domainsToDescriptions = exports.Problems = exports.Problem = exports.ArkTypeError = void 0;
const scanner_js_1 = require("../parse/string/shift/scanner.js");
const data_js_1 = require("../utils/data.js");
const domains_js_1 = require("../utils/domains.js");
const generics_js_1 = require("../utils/generics.js");
const numericLiterals_js_1 = require("../utils/numericLiterals.js");
const objectKinds_js_1 = require("../utils/objectKinds.js");
const paths_js_1 = require("../utils/paths.js");
const serialize_js_1 = require("../utils/serialize.js");
class ArkTypeError extends TypeError {
    cause;
    constructor(problems) {
        super(`${problems}`);
        this.cause = problems;
    }
}
exports.ArkTypeError = ArkTypeError;
class Problem {
    code;
    path;
    data;
    source;
    writers;
    parts;
    constructor(code, path, data, source, writers) {
        this.code = code;
        this.path = path;
        this.data = data;
        this.source = source;
        this.writers = writers;
        if (this.code === "multi") {
            this.parts = this.source;
        }
    }
    toString() {
        return this.message;
    }
    get message() {
        return this.writers.addContext(this.reason, this.path);
    }
    get reason() {
        return this.writers.writeReason(this.mustBe, new data_js_1.DataWrapper(this.data));
    }
    get mustBe() {
        return typeof this.writers.mustBe === "string"
            ? this.writers.mustBe
            : this.writers.mustBe(this.source);
    }
}
exports.Problem = Problem;
class ProblemArray extends Array {
    byPath = {};
    count = 0;
    #state;
    constructor(state) {
        super();
        this.#state = state;
    }
    mustBe(description, opts) {
        return this.add("custom", description, opts);
    }
    add(code, source, opts) {
        // copy the path to avoid future mutations affecting it
        const path = paths_js_1.Path.from(opts?.path ?? this.#state.path);
        const data = 
        // we have to check for the presence of the key explicitly since the
        // data could be undefined or null
        opts && "data" in opts
            ? opts.data
            : this.#state.data;
        const problem = new Problem(
        // avoid a bunch of errors from TS trying to discriminate the
        // problem input based on the code
        code, path, data, source, this.#state.getProblemConfig(code));
        this.addProblem(problem);
        return problem;
    }
    addProblem(problem) {
        const pathKey = `${problem.path}`;
        const existing = this.byPath[pathKey];
        if (existing) {
            if (existing.parts) {
                existing.parts.push(problem);
            }
            else {
                const problemIntersection = new Problem("multi", existing.path, existing.data, [existing, problem], this.#state.getProblemConfig("multi"));
                const existingIndex = this.indexOf(existing);
                // If existing is found (which it always should be unless this was externally mutated),
                // replace it with the new problem intersection. In case it isn't for whatever reason,
                // just append the intersection.
                this[existingIndex === -1 ? this.length : existingIndex] =
                    problemIntersection;
                this.byPath[pathKey] = problemIntersection;
            }
        }
        else {
            this.byPath[pathKey] = problem;
            this.push(problem);
        }
        this.count++;
    }
    get summary() {
        return `${this}`;
    }
    toString() {
        return this.join("\n");
    }
    throw() {
        throw new ArkTypeError(this);
    }
}
exports.Problems = ProblemArray;
const capitalize = (s) => s[0].toUpperCase() + s.slice(1);
const domainsToDescriptions = (domains) => domains.map((objectKind) => domains_js_1.domainDescriptions[objectKind]);
exports.domainsToDescriptions = domainsToDescriptions;
const objectKindsToDescriptions = (kinds) => kinds.map((objectKind) => objectKinds_js_1.objectKindDescriptions[objectKind]);
exports.objectKindsToDescriptions = objectKindsToDescriptions;
const describeBranches = (descriptions) => {
    if (descriptions.length === 0) {
        return "never";
    }
    if (descriptions.length === 1) {
        return descriptions[0];
    }
    let description = "";
    for (let i = 0; i < descriptions.length - 1; i++) {
        description += descriptions[i];
        if (i < descriptions.length - 2) {
            description += ", ";
        }
    }
    description += ` or ${descriptions[descriptions.length - 1]}`;
    return description;
};
exports.describeBranches = describeBranches;
const writeDefaultReason = (mustBe, was) => `must be ${mustBe}${was && ` (was ${was})`}`;
const addDefaultContext = (reason, path) => path.length === 0
    ? capitalize(reason)
    : path.length === 1 && (0, numericLiterals_js_1.isWellFormedInteger)(path[0])
        ? `Item at index ${path[0]} ${reason}`
        : `${path} ${reason}`;
const defaultProblemConfig = {
    divisor: {
        mustBe: (divisor) => divisor === 1 ? `an integer` : `a multiple of ${divisor}`
    },
    class: {
        mustBe: (expected) => {
            const possibleObjectKind = (0, objectKinds_js_1.getExactConstructorObjectKind)(expected);
            return possibleObjectKind
                ? objectKinds_js_1.objectKindDescriptions[possibleObjectKind]
                : `an instance of ${expected.name}`;
        },
        writeReason: (mustBe, data) => writeDefaultReason(mustBe, data.className)
    },
    domain: {
        mustBe: (domain) => domains_js_1.domainDescriptions[domain],
        writeReason: (mustBe, data) => writeDefaultReason(mustBe, data.domain)
    },
    missing: {
        mustBe: () => "defined",
        writeReason: (mustBe) => writeDefaultReason(mustBe, "")
    },
    extraneous: {
        mustBe: () => "removed",
        writeReason: (mustBe) => writeDefaultReason(mustBe, "")
    },
    bound: {
        mustBe: (bound) => `${scanner_js_1.Scanner.comparatorDescriptions[bound.comparator]} ${bound.limit}${bound.units ? ` ${bound.units}` : ""}`,
        writeReason: (mustBe, data) => writeDefaultReason(mustBe, `${data.size}`)
    },
    regex: {
        mustBe: (expression) => `a string matching ${expression}`
    },
    value: {
        mustBe: serialize_js_1.stringify
    },
    branches: {
        mustBe: (branchProblems) => (0, exports.describeBranches)(branchProblems.map((problem) => `${problem.path} must be ${problem.parts
            ? (0, exports.describeBranches)(problem.parts.map((part) => part.mustBe))
            : problem.mustBe}`)),
        writeReason: (mustBe, data) => `${mustBe} (was ${data})`,
        addContext: (reason, path) => path.length ? `At ${path}, ${reason}` : reason
    },
    multi: {
        mustBe: (problems) => "• " + problems.map((_) => _.mustBe).join("\n• "),
        writeReason: (mustBe, data) => `${data} must be...\n${mustBe}`,
        addContext: (reason, path) => path.length ? `At ${path}, ${reason}` : reason
    },
    custom: {
        mustBe: (mustBe) => mustBe
    },
    cases: {
        mustBe: (cases) => (0, exports.describeBranches)(cases)
    }
};
exports.problemCodes = (0, generics_js_1.objectKeysOf)(defaultProblemConfig);
const compileDefaultProblemWriters = () => {
    const result = {};
    let code;
    for (code of exports.problemCodes) {
        result[code] = {
            mustBe: defaultProblemConfig[code].mustBe,
            writeReason: defaultProblemConfig[code].writeReason ??
                writeDefaultReason,
            addContext: defaultProblemConfig[code].addContext ?? addDefaultContext
        };
    }
    return result;
};
exports.defaultProblemWriters = compileDefaultProblemWriters();
const compileProblemWriters = (input) => {
    if (!input) {
        return exports.defaultProblemWriters;
    }
    const result = {};
    for (const code of exports.problemCodes) {
        result[code] = {
            mustBe: input[code]?.mustBe ??
                defaultProblemConfig[code].mustBe,
            writeReason: input[code]?.writeReason ??
                defaultProblemConfig[code].writeReason ??
                input.writeReason ??
                writeDefaultReason,
            addContext: input[code]?.addContext ??
                defaultProblemConfig[code].addContext ??
                input.addContext ??
                addDefaultContext
        };
    }
    return result;
};
exports.compileProblemWriters = compileProblemWriters;
