"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.checkRequiredProp = exports.checkEntries = exports.traverse = exports.TraversalState = exports.traverseRoot = void 0;
const discriminate_js_1 = require("../nodes/discriminate.js");
const class_js_1 = require("../nodes/rules/class.js");
const divisor_js_1 = require("../nodes/rules/divisor.js");
const range_js_1 = require("../nodes/rules/range.js");
const regex_js_1 = require("../nodes/rules/regex.js");
const rules_js_1 = require("../nodes/rules/rules.js");
const domains_js_1 = require("../utils/domains.js");
const errors_js_1 = require("../utils/errors.js");
const generics_js_1 = require("../utils/generics.js");
const numericLiterals_js_1 = require("../utils/numericLiterals.js");
const paths_js_1 = require("../utils/paths.js");
const problems_js_1 = require("./problems.js");
const initializeTraversalConfig = () => ({
    mustBe: [],
    writeReason: [],
    addContext: [],
    keys: []
});
const problemWriterKeys = [
    "mustBe",
    "writeReason",
    "addContext"
];
const traverseRoot = (t, data) => {
    const state = new TraversalState(data, t);
    (0, exports.traverse)(t.flat, state);
    const result = new CheckResult(state);
    if (state.problems.count) {
        result.problems = state.problems;
    }
    else {
        for (const [o, k] of state.entriesToPrune) {
            delete o[k];
        }
        result.data = state.data;
    }
    return result;
};
exports.traverseRoot = traverseRoot;
const CheckResult = class {
    data;
    problems;
};
class TraversalState {
    data;
    type;
    path = new paths_js_1.Path();
    problems = new problems_js_1.Problems(this);
    entriesToPrune = [];
    failFast = false;
    traversalConfig = initializeTraversalConfig();
    rootScope;
    #seen = {};
    constructor(data, type) {
        this.data = data;
        this.type = type;
        this.rootScope = type.scope;
    }
    getProblemConfig(code) {
        const result = {};
        for (const k of problemWriterKeys) {
            result[k] =
                this.traversalConfig[k][0] ??
                    this.rootScope.config.codes[code][k];
        }
        return result;
    }
    traverseConfig(configEntries, node) {
        for (const entry of configEntries) {
            this.traversalConfig[entry[0]].unshift(entry[1]);
        }
        const isValid = (0, exports.traverse)(node, this);
        for (const entry of configEntries) {
            this.traversalConfig[entry[0]].shift();
        }
        return isValid;
    }
    traverseKey(key, node) {
        const lastData = this.data;
        this.data = this.data[key];
        this.path.push(key);
        const isValid = (0, exports.traverse)(node, this);
        this.path.pop();
        if (lastData[key] !== this.data) {
            lastData[key] = this.data;
        }
        this.data = lastData;
        return isValid;
    }
    traverseResolution(name) {
        const resolution = this.type.scope.resolve(name);
        const id = resolution.qualifiedName;
        // this assignment helps with narrowing
        const data = this.data;
        const isObject = (0, domains_js_1.hasDomain)(data, "object");
        if (isObject) {
            const seenByCurrentType = this.#seen[id];
            if (seenByCurrentType) {
                if (seenByCurrentType.includes(data)) {
                    // if data has already been checked by this alias as part of
                    // a resolution higher up on the call stack, it must be valid
                    // or we wouldn't be here
                    return true;
                }
                seenByCurrentType.push(data);
            }
            else {
                this.#seen[id] = [data];
            }
        }
        const lastType = this.type;
        this.type = resolution;
        const isValid = (0, exports.traverse)(resolution.flat, this);
        this.type = lastType;
        if (isObject) {
            this.#seen[id].pop();
        }
        return isValid;
    }
    traverseBranches(branches) {
        const lastFailFast = this.failFast;
        this.failFast = true;
        const lastProblems = this.problems;
        const branchProblems = new problems_js_1.Problems(this);
        this.problems = branchProblems;
        const lastPath = this.path;
        const lastKeysToPrune = this.entriesToPrune;
        let hasValidBranch = false;
        for (const branch of branches) {
            this.path = new paths_js_1.Path();
            this.entriesToPrune = [];
            if ((0, exports.checkEntries)(branch, this)) {
                hasValidBranch = true;
                lastKeysToPrune.push(...this.entriesToPrune);
                break;
            }
        }
        this.path = lastPath;
        this.entriesToPrune = lastKeysToPrune;
        this.problems = lastProblems;
        this.failFast = lastFailFast;
        return hasValidBranch || !this.problems.add("branches", branchProblems);
    }
}
exports.TraversalState = TraversalState;
const traverse = (node, state) => typeof node === "string"
    ? (0, domains_js_1.domainOf)(state.data) === node || !state.problems.add("domain", node)
    : (0, exports.checkEntries)(node, state);
exports.traverse = traverse;
const checkEntries = (entries, state) => {
    let isValid = true;
    for (let i = 0; i < entries.length; i++) {
        const [k, v] = entries[i];
        const entryAllowsData = entryCheckers[k](v, state);
        isValid &&= entryAllowsData;
        if (!isValid) {
            if (state.failFast) {
                return false;
            }
            if (i < entries.length - 1 &&
                rules_js_1.precedenceMap[k] < rules_js_1.precedenceMap[entries[i + 1][0]]) {
                // if we've encountered a problem, there is at least one entry
                // remaining, and the next entry is of a higher precedence level
                // than the current entry, return immediately
                return false;
            }
        }
    }
    return isValid;
};
exports.checkEntries = checkEntries;
const checkRequiredProp = (prop, state) => {
    if (prop[0] in state.data) {
        return state.traverseKey(prop[0], prop[1]);
    }
    state.problems.add("missing", undefined, {
        path: state.path.concat(prop[0]),
        data: undefined
    });
    return false;
};
exports.checkRequiredProp = checkRequiredProp;
const createPropChecker = (kind) => (props, state) => {
    let isValid = true;
    const remainingUnseenRequired = { ...props.required };
    for (const k in state.data) {
        if (props.required[k]) {
            isValid = state.traverseKey(k, props.required[k]) && isValid;
            delete remainingUnseenRequired[k];
        }
        else if (props.optional[k]) {
            isValid = state.traverseKey(k, props.optional[k]) && isValid;
        }
        else if (props.index && numericLiterals_js_1.wellFormedIntegerMatcher.test(k)) {
            isValid = state.traverseKey(k, props.index) && isValid;
        }
        else if (kind === "distilledProps") {
            if (state.failFast) {
                // If we're in a union (i.e. failFast is enabled) in
                // distilled mode, we need to wait to prune distilled keys
                // to avoid mutating data based on a branch that will not be
                // included in the final result. Instead, we push the object
                // and key to state to handle after traversal is complete.
                state.entriesToPrune.push([state.data, k]);
            }
            else {
                // If we're not in a union, we can safely distill right away
                delete state.data[k];
            }
        }
        else {
            isValid = false;
            state.problems.add("extraneous", state.data[k], {
                path: state.path.concat(k)
            });
        }
        if (!isValid && state.failFast) {
            return false;
        }
    }
    const unseenRequired = Object.keys(remainingUnseenRequired);
    if (unseenRequired.length) {
        for (const k of unseenRequired) {
            state.problems.add("missing", undefined, {
                path: state.path.concat(k)
            });
        }
        return false;
    }
    return isValid;
};
const entryCheckers = {
    regex: regex_js_1.checkRegex,
    divisor: divisor_js_1.checkDivisor,
    domains: (domains, state) => {
        const entries = domains[(0, domains_js_1.domainOf)(state.data)];
        return entries
            ? (0, exports.checkEntries)(entries, state)
            : !state.problems.add("cases", (0, problems_js_1.domainsToDescriptions)((0, generics_js_1.objectKeysOf)(domains)));
    },
    domain: (domain, state) => (0, domains_js_1.domainOf)(state.data) === domain ||
        !state.problems.add("domain", domain),
    bound: range_js_1.checkBound,
    optionalProp: (prop, state) => {
        if (prop[0] in state.data) {
            return state.traverseKey(prop[0], prop[1]);
        }
        return true;
    },
    // these checks work the same way, the keys are only distinct so that
    // prerequisite props can have a higher precedence
    requiredProp: exports.checkRequiredProp,
    prerequisiteProp: exports.checkRequiredProp,
    indexProp: (node, state) => {
        if (!Array.isArray(state.data)) {
            state.problems.add("class", Array);
            return false;
        }
        let isValid = true;
        for (let i = 0; i < state.data.length; i++) {
            isValid = state.traverseKey(`${i}`, node) && isValid;
            if (!isValid && state.failFast) {
                return false;
            }
        }
        return isValid;
    },
    branches: (branches, state) => state.traverseBranches(branches),
    switch: (rule, state) => {
        const dataAtPath = (0, paths_js_1.getPath)(state.data, rule.path);
        const caseKey = (0, discriminate_js_1.serializeCase)(rule.kind, dataAtPath);
        if ((0, generics_js_1.hasKey)(rule.cases, caseKey)) {
            return (0, exports.checkEntries)(rule.cases[caseKey], state);
        }
        const caseKeys = (0, generics_js_1.objectKeysOf)(rule.cases);
        const missingCasePath = state.path.concat(rule.path);
        const caseDescriptions = rule.kind === "value"
            ? caseKeys
            : rule.kind === "domain"
                ? (0, problems_js_1.domainsToDescriptions)(caseKeys)
                : rule.kind === "class"
                    ? (0, problems_js_1.objectKindsToDescriptions)(caseKeys)
                    : (0, errors_js_1.throwInternalError)(`Unexpectedly encountered rule kind '${rule.kind}' during traversal`);
        state.problems.add("cases", caseDescriptions, {
            path: missingCasePath,
            data: dataAtPath
        });
        return false;
    },
    alias: (name, state) => state.traverseResolution(name),
    class: class_js_1.checkClass,
    narrow: (narrow, state) => {
        const lastProblemsCount = state.problems.count;
        const result = narrow(state.data, state.problems);
        if (!result && state.problems.count === lastProblemsCount) {
            state.problems.mustBe(narrow.name ? `valid according to ${narrow.name}` : "valid");
        }
        return result;
    },
    config: ({ config, node }, state) => state.traverseConfig(config, node),
    value: (value, state) => state.data === value || !state.problems.add("value", value),
    morph: (morph, state) => {
        const out = morph(state.data, state.problems);
        if (state.problems.length) {
            return false;
        }
        if (out instanceof problems_js_1.Problem) {
            // if a problem was returned from the morph but not added, add it
            state.problems.addProblem(out);
            return false;
        }
        if (out instanceof CheckResult) {
            if (out.problems) {
                for (const problem of out.problems) {
                    state.problems.addProblem(problem);
                }
                return false;
            }
            state.data = out.data;
            return true;
        }
        state.data = out;
        return true;
    },
    distilledProps: createPropChecker("distilledProps"),
    strictProps: createPropChecker("strictProps")
};
